/*
Copyright 2018 Michael Hansen

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

#include <Python.h>
#include <stdio.h>
#include <string>
#include <vector>

/* Not declared via Python.h */
extern "C" {
extern DL_IMPORT(void) PyMarshal_WriteObjectToFile(PyObject *, FILE *);
}

static void str_replace(std::string &str, const char *from, const char *to)
{
    std::string::size_type it = 0;
    const size_t from_size = strlen(from);
    const size_t to_size = strlen(to);
    for ( ;; ) {
        it = str.find(from, it);
        if (it == std::string::npos)
            break;
        str.replace(it, from_size, to, to_size);
        it += to_size;
    }
}

static void wr_long(FILE *f, long x)
{
    putc((char)(x         & 0xff), f);
    putc((char)((x >> 8)  & 0xff), f);
    putc((char)((x >> 16) & 0xff), f);
    putc((char)((x >> 24) & 0xff), f);
}

/* This is essentially a C++ port of py_compile.py */
static bool compile(const char *file, const char *cfile = NULL,
                    const char *dfile = NULL)
{
    FILE *f = fopen(file, "rb");
    if (f == NULL) {
        fprintf(stderr, "Failed to open %s for reading\n", file);
        return false;
    }

    struct stat py_stat;
    long timestamp = 0;
    if (fstat(fileno(f), &py_stat) == 0 || stat(file, &py_stat) == 0) {
        timestamp = (long)py_stat.st_ctime;
    }

    std::string codestring;
    char cbuf[512];
    for ( ;; ) {
        size_t rc = fread(cbuf, 1, sizeof(cbuf), f);
        if (rc == 0)
            break;
        codestring.append(cbuf, rc);
    }
    fclose(f);

    str_replace(codestring, "\r\n", "\n");
    str_replace(codestring, "\r", "\n");

    if (codestring.size() && codestring.at(codestring.size() - 1) != '\n')
        codestring.push_back('\n');

    PyCompilerFlags cf;
    cf.cf_flags = 0;
    if (dfile == NULL)
        dfile = file;
    PyEval_MergeCompilerFlags(&cf);
    PyObject *codeobject = Py_CompileStringFlags((char *)codestring.c_str(),
                                                 (char *)dfile,
                                                 Py_file_input, &cf);
    if (codeobject == NULL) {
        // TODO: Print the traceback, but without using any python...
        fprintf(stderr, "Syntax error in %s\n", file);
        return false;
    }

    std::string cfile_s;
    if (cfile == NULL) {
        cfile_s = file;
        cfile_s.push_back(Py_OptimizeFlag ? 'o' : 'c');
    } else {
        cfile_s = cfile;
    }

    FILE *fc = fopen(cfile_s.c_str(), "wb");
    fwrite("\0\0\0\0", 4, 1, fc);
    wr_long(fc, timestamp);
    PyMarshal_WriteObjectToFile(codeobject, fc);
    fflush(fc);
    fseek(fc, 0, SEEK_SET);
    wr_long(fc, PyImport_GetMagicNumber());
    fclose(fc);

    Py_DECREF(codeobject);

    return true;
}

static void show_help(int exitcode)
{
    puts("Usage:  compyle22 [options] source.py [source2.py ...]");
    puts("");
    puts("Options:");
    puts("  -h  --help  Show this help");
    puts("  -O          Optimize compiled files (.pyo)");

    exit(exitcode);
}

int main(int argc, char *argv[])
{
    std::vector<std::string> files;

    for (int i = 1; i < argc; ++i) {
        if (strcmp(argv[i], "-O") == 0) {
            Py_OptimizeFlag = 1;
        } else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {
            show_help(0);
        } else if (argv[i][0] == '-') {
            fprintf(stderr, "Unsupported option '%s'\n\n", argv[i]);
            show_help(1);
        } else {
            files.push_back(argv[i]);
        }
    }

    if (files.empty())
        show_help(1);

    Py_NoSiteFlag = 1;
    Py_Initialize();

    std::vector<std::string>::iterator fi;
    int nfailed = 0;
    for (fi = files.begin(); fi != files.end(); ++fi) {
        if (!compile(fi->c_str()))
            ++nfailed;
    }

    Py_Finalize();

    if (nfailed != 0) {
        fprintf(stderr, "%d files failed to compile\n", nfailed);
        exit(1);
    }

    return 0;
}
